#define __TCP_SESSION_C__

#include "TCP_Session.h"
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <pthread.h>

#include "main_data.h"
#include "main_mgr_proto.h"

#include "ccs_manager.h"

#include "utils.h"

/* we have to mutex communication socket with ccs_manager */
static int spair_main;
static pthread_mutex_t spair_main_mux;

/* REGOLA MUTEX: 
 * 1. lock ccs list
 * 2. lock ccs socket pair
 * 3. unlock ccs list
 * 4. unlock ccs socket pair
 * ECCEZIONE in TCP_Socket:
 * 1. lock spair_main
 * 2. unlock spair_main
 * 3. lock ccs list
 * 4. unlock ccs list
 * ECCEZIONE in CloseCCS:
 * 1. lock ccs list
 * 2. lock spair_main
 * 3. unlock spair_main
 * 4. unlock ccs_list
 */

int Init_TCP_Session_Module(void *ptr){
pthread_attr_t tattr;
pthread_t thrid;
int* sp;
struct linger linopt;
socklen_t optlen = sizeof(struct linger);

	/* create data structures */
	initccslist();
	acquireccslist();

	/* create socket pair for communication */
	sp=malloc(2*sizeof(int));
	if(!sp){
		perror("alloc socketpair");
		exit(-1);
	}
	if((socketpair(AF_UNIX,SOCK_STREAM,0, sp)) != 0){
		perror("socketpair");
		exit(-1);
	}
	linopt.l_onoff = 1;
	linopt.l_linger = LINGER_TIMEOUT;
	if( setsockopt(sp[0], SOL_SOCKET, SO_LINGER, &linopt, optlen) ){
		perror("setting linger on socketpair\n");
		exit(-1);
	}
	if( setsockopt(sp[1], SOL_SOCKET, SO_LINGER, &linopt, optlen) ){
		perror("setting linger on socketpair\n");
		exit(-1);
	}
	spair_main = sp[0];
	pthread_mutex_init(&spair_main_mux, NULL);

	/* launch ccs_manager */
	pthread_attr_init(&tattr);
	pthread_attr_setdetachstate(&tattr, PTHREAD_CREATE_DETACHED);
	if((pthread_create(&thrid, &tattr, ccs_manager, sp)) != 0){
		perror("creating thread");
		exit(-1);
	}
	pthread_attr_destroy(&tattr);

	releaseccslist();

	return 1;

}

int Close_TCP_Session_Module(void *ptr){
	
	acquireccslist();

	pthread_mutex_lock(&spair_main_mux);
	if( mcm_putrequest(spair_main, CLOSECCS, NULL, NULL, -1, NULL, 0) == -1 ){
		dbg_printf("qualcosa e' andato terribilmente storto\n");
		exit(-1);
	}
	close(spair_main);
	pthread_mutex_unlock(&spair_main_mux);
	pthread_mutex_destroy(&spair_main_mux);

	releaseccslist();
	destroyccslist();
	
	return 1;
}

int Socket(int domain, int type, int protocol){
	return(socket(domain,type,protocol));
}

int TCP_Session_IPv4_Socket(void){
struct mcm_msg_reply* msg;
int ccsfd;
	pthread_mutex_lock(&spair_main_mux);
	if( mcm_putrequest(spair_main, SOCKT, NULL, NULL, -1, NULL, 0)
			== -1 ){
		dbg_printf("qualcosa e' andato terribilmente storto\n");
		exit(-1);
	}
	msg = mcm_getreply(spair_main);
	pthread_mutex_unlock(&spair_main_mux);

	ccsfd = msg->retval; /* ccs_mgr doesn't return -1 */

	acquireccslist();
	addccslist(ccsfd);
	releaseccslist();

	free(msg);

	return ccsfd;
}

int SetsockoptReuseAddr(int s){
struct mcm_msg_reply* msg;
pthread_mutex_t* fdmux;
int OptVal = 1;
int myerrno = EBADF;

	acquireccslist();
	
	if( (fdmux=isinccslist(s)) != NULL ){
		
		pthread_mutex_lock(fdmux);
		releaseccslist();
		
		if( mcm_putrequest(s, STRADR, NULL, NULL, -1, NULL, 0) == -1 ){
			dbg_printf("ccs socket morto\n");
			acquireccslist();
			delccslist(s);
			releaseccslist();
		}	
		msg = mcm_getreply(s);
		pthread_mutex_unlock(fdmux);

		if( msg->retval == 1){
			myerrno = msg->myerrno;
			free(msg);
			releaseccslist();
			errno = myerrno;
			return 1;
		}
		free(msg);
		
		return 0;
	} else {
		releaseccslist();
		return(!setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *)&OptVal, sizeof(OptVal)));
	}
}

int GetsockoptReuseAddr(int s, int *pFlag){
struct mcm_msg_reply* msg;
pthread_mutex_t* fdmux;
socklen_t pFlagLen = sizeof(int);
//int myerrno = EBADF;

	acquireccslist();
	
	if( (fdmux=isinccslist(s)) != NULL ){

		pthread_mutex_lock(fdmux);
		releaseccslist();

		if( mcm_putrequest(s, GTRADR, pFlag, NULL, -1, NULL, 0) == -1 ){
			dbg_printf("ccs socket morto\n");
			acquireccslist();
			delccslist(s);
			releaseccslist();
		}
		msg = mcm_getreply(s);
		pthread_mutex_unlock(fdmux);

		free(msg);

		return 1;
	} else {
		releaseccslist();
		return(getsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *)pFlag, &pFlagLen));
	}

}

int Bind(int  sockfd, struct sockaddr *my_addr, socklen_t addrlen){
struct mcm_msg_reply* msg;
pthread_mutex_t* fdmux;
int myerrno = EBADF;

	acquireccslist();
	
	if( (fdmux=isinccslist(sockfd)) != NULL ){

		pthread_mutex_lock(fdmux);
		releaseccslist();

		if( mcm_putrequest(sockfd, BIND, NULL, my_addr, -1, NULL, 0) == -1 ){
			dbg_printf("ccs socket morto\n");
			acquireccslist();
			delccslist(sockfd);
			releaseccslist();
		}
		msg = mcm_getreply(sockfd);
		pthread_mutex_unlock(fdmux);
	
		if( msg->retval == -1){
			myerrno = msg->myerrno;
			free(msg);
			errno = myerrno;
			return -1;
		}
		free(msg);

		return 0;
	} else {
		releaseccslist();
		return(bind(sockfd,my_addr,addrlen));
	}
}

int Listen(int s, int backlog){
struct mcm_msg_reply* msg;
pthread_mutex_t* fdmux;
int myerrno = EBADF;

	acquireccslist();
	
	if( (fdmux=isinccslist(s)) != NULL ){
		
		pthread_mutex_lock(fdmux);
		releaseccslist();

		if( mcm_putrequest(s, LISTEN, NULL, NULL, backlog, NULL, 0) == -1 ){
			dbg_printf("ccs socket morto\n");
			acquireccslist();
			delccslist(s);
			releaseccslist();
		}
		msg = mcm_getreply(s);
		pthread_mutex_unlock(fdmux);

		if( msg->retval == -1){
			myerrno = msg->myerrno;
			free(msg);
			errno = myerrno;
			return -1;
		}
		free(msg);
		
		return 0;
	} else {
		releaseccslist();
		return(listen(s,backlog));
	}
}

int Accept(int s,  struct  sockaddr  *addr,  socklen_t *addrlen){
struct mcm_msg_reply* msg;
pthread_mutex_t* fdmux;
int myerrno= EBADF, newccsfd;

	acquireccslist();
	
	if( (fdmux=isinccslist(s)) != NULL ){

		pthread_mutex_lock(fdmux);
		releaseccslist();

		if( mcm_putrequest(s, ACCEPT, NULL, addr, -1, NULL, 0) == -1 ){
			dbg_printf("ccs socket morto\n");
			acquireccslist();
			delccslist(s);
			releaseccslist();
		}
		msg = mcm_getreply(s); /* wait here */
		pthread_mutex_unlock(fdmux);
		
		if( msg->retval == -1){
			myerrno = msg->myerrno;
			free(msg);
			errno = myerrno;
			return -1;
		}
		newccsfd = msg->retval;
		
		acquireccslist();
		addccslist(newccsfd);
		releaseccslist();
		
		free(msg);

		return newccsfd;
	} else {
		releaseccslist();
		return(accept(s,addr,addrlen));
	}

}

int Connect(int sockfd, const struct sockaddr *serv_addr, socklen_t addrlen){
struct mcm_msg_reply* msg;
pthread_mutex_t* fdmux;
int myerrno = EBADF;

	acquireccslist();
	
	if( (fdmux=isinccslist(sockfd)) != NULL ){
		
		pthread_mutex_lock(fdmux);
		releaseccslist();

		if( mcm_putrequest(sockfd, CONNECT, NULL,
				(struct sockaddr*)serv_addr, -1, NULL, 0) == -1 ){
			dbg_printf("ccs socket morto\n");
			acquireccslist();
			delccslist(sockfd);
			releaseccslist();
		}
		msg = mcm_getreply(sockfd);
		pthread_mutex_unlock(fdmux);

		if( msg->retval == -1){
			myerrno = msg->myerrno;
			free(msg);
			errno = myerrno;
			return -1;
		}

		free(msg);
		
		return 0;
	} else {
		releaseccslist();
		return(connect(sockfd,serv_addr,addrlen));
	}

}

ssize_t Read(int fd, void *buf, size_t count){
struct mcm_msg_reply* msg;
pthread_mutex_t* fdmux;
int myerrno = EBADF;
size_t retval;

	acquireccslist();
    
	if( (fdmux=isinccslist(fd)) != NULL ){

		pthread_mutex_lock(fdmux);
		releaseccslist();

		if( mcm_putrequest(fd, READ, NULL, NULL, -1, buf, count) == -1 ){
			dbg_printf("ccs socket morto\n");
			acquireccslist();
			delccslist(fd);
			releaseccslist();
		}
		msg = mcm_getreply(fd); /* block here */
		pthread_mutex_unlock(fdmux);
		
		if( msg->retval == -1){
			myerrno = msg->myerrno;
			free(msg);
			errno = myerrno;
			return -1;
		}
		retval = (size_t)msg->retval;
		free(msg);
		
		return retval;
	} else {
		releaseccslist();
		return(read(fd,buf,count));
	}

}

ssize_t Write(int fd, const void *buf, size_t count) {
struct mcm_msg_reply* msg;
pthread_mutex_t* fdmux;
int myerrno = EBADF, retval;

	acquireccslist();
    

	if( (fdmux=isinccslist(fd)) != NULL ){

		pthread_mutex_lock(fdmux);
		releaseccslist();

		if( mcm_putrequest(fd, WRITE, NULL, NULL, -1, (void *)buf, count) == -1 ){
			dbg_printf("ccs socket morto\n");
			acquireccslist();
			delccslist(fd);
			releaseccslist();
		}
		msg = mcm_getreply(fd);
		pthread_mutex_unlock(fdmux);

		if( msg->retval == -1){ /* ccs_manager return -1 or count */
			myerrno = msg->myerrno;
			free(msg);
			errno = myerrno;
			return -1;
		}
		retval = msg->retval;
		free(msg);
		
		return retval;
	} else {
		releaseccslist();
		return(write(fd,buf,count));
	}

}

int Close(int fd){
pthread_mutex_t* fdmux;

	acquireccslist();
	
	if( (fdmux=isinccslist(fd)) != NULL ){

		pthread_mutex_lock(fdmux);
		releaseccslist();

		mcm_putrequest(fd, CLOSE, NULL, NULL, -1, NULL, 0);
		pthread_mutex_unlock(fdmux);

		acquireccslist();
		delccslist(fd);
		releaseccslist();

		return 0;
	} else {
		releaseccslist();
		return(close(fd));
	}

}


