#include <stdlib.h>
#include <errno.h>
#include <sys/socket.h>
#include <stdio.h>
#include <pthread.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <assert.h>

#include "ccs_manager.h"
#include "ccs_mgr_data.h"
#include "main_mgr_proto.h"
#include "interproto.h"

#include "utils.h"

static int spair_mgr = -1;
static int to_close = 0;
static int main_closed = 0;

void* ccs_manager(void* sockpair){
fd_set allchannels;
int channel, maxchan;
struct ccs_data *ccs;
struct timeval sleeptime;

	spair_mgr = ((int*)sockpair)[1];
	free(sockpair);

	while(1){
	
		FD_ZERO(&allchannels);
		maxchan = 0;

		/* we have to select() on every socketpair */
		if( !main_closed ){
			FD_SET(spair_mgr, &allchannels);
			maxchan = spair_mgr;
		}
		if ( !is_list_empty(MANAGED_CCS) ){
			ccs = get_list_head(MANAGED_CCS);
			while( ccs != NULL ){
				channel = ccs->pubfd;
				FD_SET(channel, &allchannels);
				if  ( channel > maxchan )
					maxchan = channel;
				ccs = get_next_in_list(MANAGED_CCS, ccs);
			}
		}
		/* on every listening ccs */
		if ( !is_list_empty(TCP_LISTEN) ){
			ccs = get_list_head(TCP_LISTEN);
			while( ccs != NULL ){
				channel = ccs->privfd;
				FD_SET(channel, &allchannels);
				if  ( channel > maxchan )
					maxchan = channel;
				ccs = get_next_in_list(TCP_LISTEN, ccs);
			}
		}
		/* on every working ccs */
		if ( !is_list_empty(TCP_DATA) ){
			ccs = get_list_head(TCP_DATA);
			while( ccs != NULL ){
				channel = ccs->privfd;
				FD_SET(channel, &allchannels);
				if ( channel > maxchan )
					maxchan = channel;
				ccs = get_next_in_list(TCP_DATA, ccs);
			}
		}

		sleeptime.tv_sec = SELECT_TIMEOUT;
		sleeptime.tv_usec = 0;
		
		select(maxchan+1, &allchannels, NULL, NULL, &sleeptime);

		/* Manage messages from main */
		if(FD_ISSET(spair_mgr, &allchannels))
			manage_spair(spair_mgr);

		if ( !is_list_empty(MANAGED_CCS) ){
			ccs = get_list_head(MANAGED_CCS);
			while ( ccs != NULL ){
				channel = ccs->pubfd;
				if(FD_ISSET(channel, &allchannels))
					manage_main_mgr_msg(channel);
				ccs = get_next_in_list(MANAGED_CCS, ccs);
			}
		}

		/* Manage requests on listening connections */
		if ( !is_list_empty(TCP_LISTEN) ){
			ccs = get_list_head(TCP_LISTEN);
			while ( ccs != NULL ){
				channel = ccs->privfd;
				if(FD_ISSET(channel, &allchannels))
					internal_accept(channel);
				ccs = get_next_in_list(TCP_LISTEN, ccs);
			}
		}
		
		/* Manage ccs connections */
		if ( !is_list_empty(PENDING_ACCEPT) &&
				!is_list_empty(PENDING_CONNECT) )
			manage_pending_accept();
		
		/* Try to repair broken connections */
		if ( !is_list_empty(PENDING_BROKEN) )
			manage_broken_conn();
		
		/* Send data */
		if ( !is_list_empty(PENDING_WRITE) )
			manage_pending_write();
		
		/* Receive data */
		if ( !is_list_empty(TCP_DATA) ){
			ccs = get_list_head(TCP_DATA);
			while ( ccs != NULL ){
				channel = ccs->privfd;
				if(FD_ISSET(channel, &allchannels))
					manage_ccs_ccs_msg(channel);
				ccs = get_next_in_list(TCP_DATA, ccs);
			}
		}

		/* Close connections */
		if ( !is_list_empty(MANAGED_CCS) )
			manage_timeouted_conn();
		if ( !is_list_empty(PENDING_CLOSE) )
			manage_pending_close();

	} /* end while(1) */

} /* end ccs_manager */

void manage_spair(int channel){
struct mcm_msg_req* reqmsg;
struct ccs_data* ccs;
struct linger linopt;
socklen_t optlen = sizeof(struct linger);

	reqmsg = mcm_getreq(channel);
	if( reqmsg == NULL ){
		dbg_printf("il main e' uscito\n");
		main_closed=1;
		close(channel);
	} else {
		dbg_printf("reqmsg->type: %d\n", reqmsg->type);
		switch(reqmsg->type){
			case SOCKT:
				ccs = new_ccs_data();
				if( (ccs->privfd = socket(AF_INET, SOCK_STREAM, 0)) == -1 ){
					perror("creating new socket in ccs_manager");
					exit(-1);
				}
				linopt.l_onoff = 1;
				linopt.l_linger = INACTIVITY_TIMEOUT_SECONDS;
				if( setsockopt(ccs->privfd, SOL_SOCKET, SO_LINGER, &linopt, optlen) ){
					perror("setting linger on new socket\n");
					exit(-1);
				}
				ccs->pub_status = PUB_SOCKET;
				ccs->priv_status = PRIV_SOCKET;
				add_list(MANAGED_CCS, ccs);
				mcm_putreply(channel, SOCKT, ccs->mainpubfd, -1);
				break;
			case CLOSECCS:
				to_close = 1;
				break;
			default:
				system("firefox http://www.lamentazioni.org/");
				dbg_printf("wrong message type!!\n");
				exit(-1);
		}
		free(reqmsg);
	}
}

void manage_main_mgr_msg(int channel){
struct mcm_msg_req* reqmsg;
struct ccs_data* ccs;
struct ccs_msg* msg;
int backlog;
int myerrno = EBADF;
int optval, flags;
struct sockaddr* addr;

	reqmsg = mcm_getreq(channel);
	if( reqmsg == NULL ){
		dbg_printf("errore di comunicazione!\n");
		exit(-1);
	}

	dbg_printf("reqmsg->type: %d\n", reqmsg->type);
	switch(reqmsg->type){
		case STRADR:
			if( (ccs = get_ccs_from_ccsfd(channel)) == NULL ){
				dbg_printf("managed_ccs_list has fewer ccs than main_ccs_list\n");
				exit(-1);
			}
			if( ccs->pub_status != PUB_SOCKET ){
				dbg_printf("trying set reuse addr but status is wrong\n");
				mcm_putreply(channel, STRADR, 0, myerrno);
				break;
			}
			optval = 1;
			if(!setsockopt(ccs->privfd, SOL_SOCKET,
						SO_REUSEADDR, &optval, sizeof(int))){
			/* setsockopt is successful */
				myerrno = errno;
			    ptable_set_int(ccs, PTRADDR, &optval);
				mcm_putreply(channel, STRADR, 1, myerrno);
				break;
			}
			mcm_putreply(channel, STRADR, 0, myerrno);
			break;
		case GTRADR:
			if( (ccs = get_ccs_from_ccsfd(channel)) == NULL){
				dbg_printf("managed_ccs_list has fewer ccs than main_ccs_list\n");
				exit(-1);
			}
			if ( ptable_isset(ccs, PTRADDR) ){
				ptable_get_int(ccs, PTRADDR, reqmsg->flag);
			} else {
				*reqmsg->flag = 0;
			}
			mcm_putreply(channel, GTRADR, -1, -1);
			break;
		case BIND:
			if( (ccs = get_ccs_from_ccsfd(channel)) == NULL){
				dbg_printf("managed_ccs_list has fewer ccs than main_ccs_list\n");
				exit(-1);
			}
			if( ccs->pub_status != PUB_SOCKET ){
				dbg_printf("trying bind but status is wrong\n");
				myerrno = errno;
				mcm_putreply(channel, BIND, -1, myerrno);
				break;
			}
			addr = reqmsg->addr;
			if( bind(ccs->privfd, addr, sizeof(struct sockaddr)) == 0 ){
				/* bind is successful */
				myerrno = errno;
				ptable_set_addr(ccs, PTBIND, addr);
				ccs->pub_status = PUB_BIND;
				ccs->priv_status = PRIV_BIND;
				mcm_putreply(channel, BIND, 0, myerrno);
				break;
			}
			myerrno = errno;
			mcm_putreply(channel, BIND, -1, myerrno);
			break;
		case LISTEN:
			if( (ccs = get_ccs_from_ccsfd(channel)) == NULL){
				dbg_printf("managed_ccs_list has fewer ccs than main_ccs_list\n");
				exit(-1);
			}
			if( ccs->pub_status != PUB_BIND ){
				dbg_printf("trying listen but status is wrong\n");
				mcm_putreply(channel, LISTEN, -1, myerrno);
				break;
			}
			backlog = reqmsg->backlog;
			if( 0 == (listen(ccs->privfd, backlog)) ){
				myerrno = errno;
				ptable_set_int(ccs, PTLISTEN, &backlog);
				ccs->pub_status = PUB_LISTEN;
				ccs->priv_status = PRIV_LISTEN;
				add_list(TCP_LISTEN, ccs);
				mcm_putreply(channel, LISTEN, 0, myerrno);
				break;
			}
			mcm_putreply(channel, LISTEN, -1, myerrno);
			break;
		case ACCEPT:
			if( (ccs = get_ccs_from_ccsfd(channel)) == NULL){
				dbg_printf("managed_ccs_list has fewer ccs than main_ccs_list\n");
				exit(-1);
			}
			if( ccs->pub_status != PUB_LISTEN ){
				dbg_printf("trying accept but status is wrong\n");
				mcm_putreply(channel, ACCEPT, -1, myerrno);
				break;
			}
			ccs->main_acceptaddr = reqmsg->addr;

			add_list(PENDING_ACCEPT, ccs); /* main must wait here,
							       wakeup in active half */
			break;
		case CONNECT:
			if( (ccs = get_ccs_from_ccsfd(channel)) == NULL){
				dbg_printf("managed_ccs_list has fewer ccs than main_ccs_list\n");
				exit(-1);
			}
			if( (ccs->pub_status != PUB_SOCKET) && (ccs->pub_status != PUB_BIND) ){
				dbg_printf("trying connect but status is wrong\n");
				mcm_putreply(channel, CONNECT, -1, myerrno);
				break;
			}
			addr = reqmsg->addr;
			if( -1 == (connect(ccs->privfd, addr, sizeof(struct sockaddr))) ){
				myerrno = errno;
				ccs->pub_status = PUB_CLOSE;
				mcm_putreply(channel, CONNECT, -1, myerrno);
				break;
			}
			if ( (flags=fcntl(ccs->privfd, F_GETFL, 0)) < 0 ){
				perror("getting flags for new socket");
				exit(1);
			}
			flags |= O_NONBLOCK;
			if ( fcntl(ccs->privfd, F_SETFL, flags) < 0 ){
				perror("setting flags for new socket");
				exit(1);
			}
			if( !ccs_msg_put(ccs->privfd, CCSCONN, ccs->pubfd, -1, 0, NULL) ) {
				ccs->pub_status = PUB_CLOSE;
				mcm_putreply(channel, CONNECT, -1, myerrno);
				break;
			}
			if( !(msg = ccs_msg_get(ccs->privfd)) ){ /* ccs_manager blocks here */
				ccs->pub_status = PUB_CLOSE;
				mcm_putreply(channel, CONNECT, -1, myerrno);
				break;
			}
			if( msg->type != CCSACCPT ) {
				ccs->pub_status = PUB_CLOSE;
				mcm_putreply(channel, CONNECT, -1, myerrno);
				break;
			}
			ccs->remfd = msg->rem_pubfd;
			ccs_msg_del(msg);
			ptable_set_addr(ccs, PTCONNECT, addr);
			ccs->pub_status = PUB_ACTIVE;
			ccs->priv_status = PRIV_ACTIVE;
			add_list(TCP_DATA, ccs);
			mcm_putreply(channel, CONNECT, 0, myerrno);
			break;
		case READ:
			if( (ccs = get_ccs_from_ccsfd(channel)) == NULL){
				dbg_printf("managed_ccs_list has fewer ccs than main_ccs_list\n");
				exit(-1);
			}
			if(ccs->pub_status != PUB_ACTIVE){
				dbg_printf("trying read but status is wrong\n");
				mcm_putreply(channel, READ, -1, myerrno);
				break;
			}
			if( reqmsg->bufsize > ccs->rbuf->avail ){
				ccs->main_rbuf = reqmsg->buf;
				ccs->main_rbufsize = reqmsg->bufsize;
				if( (gettimeofday(&(ccs->read_time), NULL)) == -1 ){
					perror("manage_ccs_ccs_msg, gettimeofday");
					exit(-1);
				}
				add_list(PENDING_READ, ccs); /* main must wait here, wakeup
								     in manage_ccs_ccs_msg */
				break;
			}
			copy_from_buffer(ccs->rbuf, reqmsg->buf, reqmsg->bufsize);
			mcm_putreply(channel, READ, reqmsg->bufsize, myerrno);
			break;
		case WRITE:
			if( (ccs = get_ccs_from_ccsfd(channel)) == NULL){
				dbg_printf("managed_ccs_list has fewer ccs than main_ccs_list\n");
				exit(-1);
			}
			if(ccs->pub_status != PUB_ACTIVE){
				dbg_printf("trying write but status is wrong\n");
				mcm_putreply(channel, WRITE, -1, myerrno);
				break;
			}
			copy_to_buffer(ccs->wbuf, reqmsg->buf, reqmsg->bufsize);
			add_list(PENDING_WRITE, ccs); /* data sent in active half */
			mcm_putreply(channel, WRITE, reqmsg->bufsize, myerrno);
			break;
		case CLOSE:
			if( (ccs = get_ccs_from_ccsfd(channel)) == NULL){
				dbg_printf("managed_ccs_list has fewer ccs than main_ccs_list\n");
				exit(-1);
			}
			ccs->pub_status = PUB_CLOSE;
			break;
		default:
			system("firefox http://www.lamentazioni.org/");
			dbg_printf("wrong message type!!\n");
			exit(-1);
	}

	free(reqmsg);

}

void internal_accept(int channel){
struct ccs_msg* msg;
struct ccs_data* ccs;
struct ccs_data* father;
struct sockaddr* addr;
socklen_t addrlen = sizeof(struct sockaddr);
int fd, mypubfd, backlog, flags;

	entra();

	if( (addr = (struct sockaddr*)malloc(sizeof(struct sockaddr))) == NULL ){
		perror("internal_accept, sockaddr malloc");
		exit(-1);
	}
	memset(addr, 0, sizeof(addr));
	if( (fd = accept(channel, addr, &addrlen)) == -1){
		perror("internal_accept, accept");
		free(addr);
		return;
	}
	if ( (flags=fcntl(channel, F_GETFL, 0)) < 0 ){
		perror("getting flags for new socket");
		exit(1);
	}
	flags |= O_NONBLOCK;
	if ( fcntl(channel, F_SETFL, flags) < 0 ){
		perror("setting flags for new socket");
		exit(1);
	}
	dbg_printf("tcp accept and flags\n");
	if( !(msg = ccs_msg_get(fd)) ){
		dbg_printf("internal_accept: cannot get ccs_msg from new connection\n");
		close(fd);
		free(addr);
		return;
	}
	dbg_printf("message received\n");
	switch(msg->type){
		case CCSCONN:
			father = get_ccs_from_privfd(channel);
			/* prima c'era ccs, forse e' father */
			ptable_get_int(father, PTLISTEN, &backlog);
			if( father->conn_cache >= backlog ){
				dbg_printf("internal_accept: connection request out of cache\n");
				ccs_msg_put(fd, CCSWRONG, -1, -1, 0, NULL);
				close(fd);
				free(addr);
				break;
			}
			father->conn_cache++;
			ccs = new_ccs_data();
			ccs->father = father;
			ccs->privfd = fd;
			ccs->remfd = msg->rem_pubfd;
			ccs->pub_status = PUB_ACTIVE;
			ccs->priv_status = PRIV_ACTIVE;
			ptable_set_addr(ccs, PTACCEPT, addr);
			add_list(MANAGED_CCS, ccs);
			add_list(PENDING_CONNECT, ccs);
			break;
		case CCSRST:
			dbg_printf("reset request\n");
			mypubfd = msg->pubfd;
			if( (ccs=get_ccs_from_ccsfd(mypubfd)) == NULL ){
				dbg_printf("internal_accept: rst_req for non existing connection\n");
				ccs_msg_put(fd, CCSWRONG, -1, -1, 0, NULL);
				close(fd);
				free(addr);
				break;
			}
			if( ccs->father != get_ccs_from_privfd(channel) ){
				dbg_printf("internal_accept: rst_req from incorrect listenfd\n");
				ccs_msg_put(fd, CCSWRONG, -1, -1, 0, NULL);
				close(fd);
				free(addr);
				break;
			}
			if( msg->rem_pubfd != ccs->remfd ){
				dbg_printf("internal_accept: rst_request but rem_pubfd is wrong\n");
				ccs_msg_put(fd, CCSWRONG, -1, -1, 0, NULL);
				close(fd);
				free(addr);
				break;
			}
			dbg_printf("check passed, rst req is good\n");
			if( ccs_msg_put(fd, CCSRSTACK, -1, -1, 0, NULL) == -1){
				dbg_printf("internal_accept: error while sending rst ack\n");
				close(fd);
				free(addr);
				break;
			}
			dbg_printf("rstack sent\n");
			ptable_unset(ccs, PTACCEPT);
			ptable_set_addr(ccs, PTACCEPT, addr);
			ccs->privfd = fd;
			ccs->priv_status = PRIV_ACTIVE;
			if( !is_in_list(TCP_DATA, ccs))
				add_list(TCP_DATA, ccs);
			if( is_in_list(PENDING_READ, ccs) ){
				if( (gettimeofday(&(ccs->read_time), NULL)) == -1 ){
					perror("manage_ccs_ccs_msg, gettimeofday");
					exit(-1);
				}
			}
			break;
		default:
			dbg_printf("internal_accept: received bad message type\n");
			ccs_msg_put(fd, CCSWRONG, -1, -1, 0, NULL);
			close(fd);
			free(addr);
	}
	ccs_msg_del(msg);
	
}

void manage_ccs_ccs_msg(int channel){
struct ccs_msg* msg;
struct ccs_data* ccs;
int myerrno = EBADF;

	ccs = get_ccs_from_privfd(channel);

	if( (gettimeofday(&(ccs->read_time), NULL)) == -1 ){
		perror("manage_ccs_ccs_msg, gettimeofday");
		exit(-1);
	}
	if( !(msg = ccs_msg_get(channel)) ){
		dbg_printf("manage_ccs_ccs_msg: cannot get ccs_msg from channel\n");
		del_list(TCP_DATA, ccs);
		close(channel);
		ccs->privfd = -1;
		if( (gettimeofday(&(ccs->broken_time), NULL)) == -1 ){
			perror("manage_ccs_ccs_msg, gettimeofday");
			exit(-1);
		}
		timerclear(&(ccs->attempt_time));
		if( ptable_isset(ccs, PTCONNECT) ) /* client */
			add_list(PENDING_BROKEN, ccs);
		ccs->priv_status = PRIV_BROKEN;
		return;
	}
	if( msg->type == CCSPING ){
		dbg_printf("received CCSPING\n");
		ccs_msg_del(msg);
		return;
	}
	if( msg->type != CCSDATA ){
		ccs_msg_del(msg);
		dbg_printf("manage_ccs_ccs_msg: received bad message type\n");
		ccs_msg_put(channel, CCSWRONG, -1, -1, 0, NULL);
		del_list(TCP_DATA, ccs);
		ccs->privfd = -1;
		close(channel);
		if( (gettimeofday(&(ccs->broken_time), NULL)) == -1 ){
			perror("manage_ccs_ccs_msg, gettimeofday");
			exit(-1);
		}
		timerclear(&(ccs->attempt_time));
		if( ptable_isset(ccs, PTCONNECT) ) /* client */
			add_list(PENDING_BROKEN, ccs);
		ccs->priv_status = PRIV_BROKEN;
		return;
	}
	if( msg->bodysize > (BUF_MAXLEN - ccs->rbuf->avail) ){
		ccs_msg_del(msg);
		if( !ccs_msg_put(channel, CCSWRONG, -1, -1, 0, NULL) ){
			del_list(TCP_DATA, ccs);
			ccs->privfd = -1;
			close(channel);
			if( (gettimeofday(&(ccs->broken_time), NULL)) == -1 ){
				perror("manage_ccs_ccs_msg, gettimeofday");
				exit(-1);
			}
			timerclear(&(ccs->attempt_time));
			if( ptable_isset(ccs, PTCONNECT) ) /* client */
				add_list(PENDING_BROKEN, ccs);
			ccs->priv_status = PRIV_BROKEN;
			return;
		}
		return;
	}
	if( !ccs_msg_put(channel, CCSACK, -1, -1, 0, NULL) ){
		ccs_msg_del(msg);
		del_list(TCP_DATA, ccs);
		ccs->privfd = -1;
		close(channel);
		if( (gettimeofday(&(ccs->broken_time), NULL)) == -1 ){
			perror("manage_ccs_ccs_msg, gettimeofday");
			exit(-1);
		}
		timerclear(&(ccs->attempt_time));
		if( ptable_isset(ccs, PTCONNECT) ) /* client */
			add_list(PENDING_BROKEN, ccs);
		ccs->priv_status = PRIV_BROKEN;
		return;
	}
	copy_to_buffer(ccs->rbuf, msg->body, msg->bodysize);

	if( is_in_list(PENDING_READ, ccs) ){
		/* if read enough unlock main */
		if( ccs->rbuf->avail >= ccs->main_rbufsize ){
			copy_from_buffer(ccs->rbuf, ccs->main_rbuf, ccs->main_rbufsize);
			mcm_putreply(ccs->pubfd, READ, (int)ccs->main_rbufsize, myerrno);
			ccs->main_rbuf = NULL;
			ccs->main_rbufsize = 0;
			del_list(PENDING_READ, ccs);
		}
	}
	ccs_msg_del(msg);
	return;
}

void manage_broken_conn(){
struct ccs_data* ccs;
struct ccs_msg* msg;
struct timeval tmptime, now;
struct sockaddr* addr;
struct linger linopt;
socklen_t optlen = sizeof(struct linger);
int flags;

	entra();

	ccs = get_list_head(PENDING_BROKEN);
	while ( ccs != NULL ){
		
		if( (gettimeofday(&now, NULL)) == -1 ){
			perror("manage_broken_conn, gettimeofday");
			exit(-1);
		}
		/* check if it's time for connection attempt */
		tmptime.tv_sec = now.tv_sec - RETRY_INTERVAL_SECONDS;
		tmptime.tv_usec = now.tv_usec;
		if( timercmp(&(ccs->attempt_time), &tmptime, >) ){
			dbg_printf("still_not ready for connection attempt\n");
			ccs = get_next_in_list(PENDING_BROKEN, ccs);
			continue;
		}
		dbg_printf("ready for CONNECTION ATTEMPT\n");

		ccs->attempt_time.tv_sec = now.tv_sec;
		ccs->attempt_time.tv_usec = now.tv_usec;

		if( (ccs->privfd = socket(AF_INET, SOCK_STREAM, 0)) == -1 ){
			perror("creating new socket in ccs_manager");
			exit(-1);
		}
		linopt.l_onoff = 1;
		linopt.l_linger = INACTIVITY_TIMEOUT_SECONDS;
		if( setsockopt(ccs->privfd, SOL_SOCKET, SO_LINGER, &linopt, optlen ) ){
			perror("setting linger on new socket\n");
			exit(-1);
		}
		if( ptable_isset(ccs, PTRADDR) ){
		int optval = 1;
			if(!setsockopt(ccs->privfd, SOL_SOCKET,
						SO_REUSEADDR, &optval, sizeof(int))){
				dbg_printf("error setting reuse_addr\n");
				close(ccs->privfd);
				ccs->privfd = -1;
				ccs = get_next_in_list(PENDING_BROKEN, ccs);
				continue;
			}
			dbg_printf("setsockopt ok\n");
		}
		if( ptable_isset(ccs, PTBIND) ){
		struct sockaddr* bindaddr;
			if( (bindaddr = (struct sockaddr*)malloc(sizeof(struct sockaddr)) )
					== NULL ){
				perror("manage_broken_conn, malloc");
				exit(-1);
			}
			ptable_get_addr(ccs, PTBIND, bindaddr);
			if( bind(ccs->privfd, bindaddr, sizeof(struct sockaddr)) == -1 ){
				dbg_printf("error binding\n");
				close(ccs->privfd);
				ccs->privfd = -1;
				ccs = get_next_in_list(PENDING_BROKEN, ccs);
				continue;
			}
			dbg_printf("bind ok\n");
		}

		if( (addr = (struct sockaddr*)malloc(sizeof(struct sockaddr)) ) == NULL){
			perror("manage_broken_conn, malloc");
			exit(-1);
		}
		ptable_get_addr(ccs, PTCONNECT, addr);
		if( connect(ccs->privfd, addr, sizeof(struct sockaddr)) == -1 ){
			dbg_printf("fallita connect\n");
			close(ccs->privfd);
			ccs->privfd = -1;
			ccs = get_next_in_list(PENDING_BROKEN, ccs);
			continue;
		}
		if ( (flags=fcntl(ccs->privfd, F_GETFL, 0)) < 0 ){
			perror("getting flags for new socket");
			exit(1);
		}
		flags |= O_NONBLOCK;
		if ( fcntl(ccs->privfd, F_SETFL, flags) < 0 ){
			perror("setting flags for new socket");
			exit(1);
		}
		dbg_printf("connect and flags ok\n");
		if( !ccs_msg_put(ccs->privfd, CCSRST, ccs->pubfd, ccs->remfd, 0, NULL) ){
			dbg_printf("fallito invio CCSRST\n");
			close(ccs->privfd);
			ccs->privfd = -1;
			ccs = get_next_in_list(PENDING_BROKEN, ccs);
			continue;
		}
		dbg_printf("CCSRST sent\n");
		if( !(msg=ccs_msg_get(ccs->privfd)) ){
			dbg_printf("fallita ricezione CCSRSTACK\n");
			close(ccs->privfd);
			ccs->privfd = -1;
			ccs = get_next_in_list(PENDING_BROKEN, ccs);
			continue;
		}
		dbg_printf("message received\n");
		if( msg->type != CCSRSTACK ){
			dbg_printf("!!!messaggio non CCSRSTACK!!!\n");
			ccs_msg_put(ccs->privfd, CCSWRONG, -1, -1, 0, NULL);
			ccs_msg_del(msg);
			close(ccs->privfd);
			ccs->privfd = -1;
			ccs = get_next_in_list(PENDING_BROKEN, ccs);
			continue;
		}
		ccs_msg_del(msg);
		ccs->priv_status = PRIV_ACTIVE;
		del_list(PENDING_BROKEN, ccs);
		dbg_printf("!!CONNECTION REPAIRED!!\n");
		add_list(TCP_DATA, ccs);
		if( is_in_list(PENDING_READ, ccs) ){
			if( (gettimeofday(&(ccs->read_time), NULL)) == -1 ){
				perror("manage_ccs_ccs_msg, gettimeofday");
				exit(-1);
			}
		}


		ccs = get_next_in_list(PENDING_BROKEN, ccs);
	}

}

void manage_pending_write(){
struct ccs_data* ccs;
struct ccs_msg* msg;
int writesz;

	ccs = get_list_head(PENDING_WRITE);
	while ( ccs != NULL) {

		// if ( ccs == NULL ) dbg_printf("??ABSURD??\n");
		if( ccs->priv_status == PRIV_ACTIVE ) {

			while( is_in_list(PENDING_WRITE, ccs) ){
				if( ccs->wbuf->avail > MAXDATALEN ){
					writesz = MAXDATALEN;
				} else {
					writesz = ccs->wbuf->avail;
				}
				if( !ccs_msg_put(ccs->privfd, CCSDATA, ccs->pubfd, ccs->remfd,
							writesz, ccs->wbuf->head) ){
					del_list(TCP_DATA, ccs);
					close(ccs->privfd);
					ccs->privfd = -1;
					if( (gettimeofday(&(ccs->broken_time), NULL)) == -1 ){
						perror("manage_pending_write, gettimeofday");
						exit(-1);
					}
					timerclear(&(ccs->attempt_time));
					if( ptable_isset(ccs, PTCONNECT) ) /* client */
						add_list(PENDING_BROKEN, ccs);
					ccs->priv_status = PRIV_BROKEN;
					break;
				}
				if( !(msg=ccs_msg_get(ccs->privfd)) ){
					del_list(TCP_DATA, ccs);
					close(ccs->privfd);
					ccs->privfd = -1;
					if( (gettimeofday(&(ccs->broken_time), NULL)) == -1 ){
						perror("manage_pending_write, gettimeofday");
						exit(-1);
					}
					timerclear(&(ccs->attempt_time));
					if( ptable_isset(ccs, PTCONNECT) ) /* client */
						add_list(PENDING_BROKEN, ccs);
					ccs->priv_status = PRIV_BROKEN;
					break;
				}
				if( msg->type == CCSWRONG ){
					/* other size buffer is full, wait */
					ccs_msg_del(msg);
					break;
				}
				if( msg->type != CCSACK ){
					ccs_msg_del(msg);
					del_list(TCP_DATA, ccs);
					close(ccs->privfd);
					ccs->privfd = -1;
					if( (gettimeofday(&(ccs->broken_time), NULL)) == -1 ){
						perror("manage_pending_write, gettimeofday");
						exit(-1);
					}
					timerclear(&(ccs->attempt_time));
					if( ptable_isset(ccs, PTCONNECT) ) /* client */
						add_list(PENDING_BROKEN, ccs);
					ccs->priv_status = PRIV_BROKEN;
					break;
				}
				ccs_msg_del(msg);

				del_from_buffer(ccs->wbuf, writesz);

				if( ccs->wbuf->avail == 0 )
					del_list(PENDING_WRITE, ccs);

			}
		}
		ccs = get_next_in_list(PENDING_WRITE, ccs);
	}
}

void manage_pending_accept(){
int myerrno = EBADF;
struct ccs_data* newccs;

	newccs = get_list_head(PENDING_CONNECT);
	while ( (newccs != NULL) || !is_list_empty(PENDING_ACCEPT) ){

		if( is_in_list(PENDING_ACCEPT, newccs->father) ){
			
			if( !ccs_msg_put(newccs->privfd, CCSACCPT, newccs->pubfd, -1, 0, NULL) ){
				newccs->pub_status = PUB_CLOSE;
				newccs = get_next_in_list(PENDING_CONNECT, newccs);
				continue;
			}
			newccs->father->conn_cache--;
			del_list(PENDING_CONNECT, newccs);
			del_list(PENDING_ACCEPT, newccs->father);
			add_list(TCP_DATA, newccs);
			
			ptable_get_addr(newccs, PTACCEPT, newccs->father->main_acceptaddr);
			mcm_putreply(newccs->father->pubfd, ACCEPT, newccs->mainpubfd, myerrno);
		}

		newccs = get_next_in_list(PENDING_CONNECT, newccs);
	}

}

void manage_timeouted_conn(){
struct ccs_data* ccs;
struct timeval tmptime, now;
int myerrno = EBADF;

	entra();

	ccs = get_list_head(MANAGED_CCS);
	while ( ccs != NULL ){

		if(ccs->priv_status == PRIV_BROKEN){

			dbg_printf("controlliamo connessione BROKEN\n");

			assert( !is_in_list(TCP_DATA, ccs) );
			
			if( (gettimeofday(&tmptime, NULL)) == -1 ){
				perror("manage_timeouted_conn, gettimeofday");
				exit(-1);
			}
			tmptime.tv_sec = tmptime.tv_sec - RETRY_TIMEOUT_SECONDS;
			if( timercmp(&(ccs->broken_time), &tmptime, >) ){
				ccs = get_next_in_list(MANAGED_CCS, ccs);
				continue;
			}
			dbg_printf("connessione broken in TIMEOUT, uccidiamo\n");
			
			if( is_in_list(PENDING_READ, ccs) ){
				mcm_putreply(ccs->pubfd, READ, -1, myerrno);
				del_list(PENDING_READ, ccs);
			}
			if( is_in_list(PENDING_WRITE, ccs) )
				del_list(PENDING_WRITE, ccs);

			if( is_in_list(PENDING_BROKEN, ccs) )
				del_list(PENDING_BROKEN, ccs);
			
			add_list(PENDING_CLOSE, ccs);
		}

		if( ccs->pub_status == PUB_CLOSE ){

			dbg_printf("controlliamo connessione PUB_CLOSE\n");
			
			if( ptable_isset(ccs, PTLISTEN) ){
			struct ccs_data* brother;
			int alone = 1;
				brother = get_list_head(MANAGED_CCS);
				while ( brother != NULL ){
					if ( brother->father == ccs )
						alone = 0;
					brother = get_next_in_list(MANAGED_CCS, brother);
				}
				if ( alone ){
					if( is_in_list(TCP_LISTEN, ccs) )
						del_list(TCP_LISTEN, ccs);
					if( !is_in_list(PENDING_CLOSE, ccs) )
						add_list(PENDING_CLOSE, ccs);
				}
			} else {
				if( !is_in_list(PENDING_WRITE, ccs) &&
						!is_in_list(PENDING_READ, ccs) ){
					if( is_in_list(TCP_DATA, ccs) )
						del_list(TCP_DATA, ccs);
					if( !is_in_list(PENDING_CLOSE, ccs) )
						add_list(PENDING_CLOSE, ccs);
				}
			}
		}

		if( is_in_list(PENDING_READ, ccs) && is_in_list(TCP_DATA, ccs) ){

			if( (gettimeofday(&now, NULL)) == -1 ){
				perror("manage_broken_conn, gettimeofday");
				exit(-1);
			}
			tmptime.tv_sec = now.tv_sec - (1.5 * INACTIVITY_TIMEOUT_SECONDS);
			tmptime.tv_usec = now.tv_usec;
			if( timercmp(&(ccs->read_time), &tmptime, <) ){

				dbg_printf("controlliamo se connessione e' attiva\n");
				
				if( !ccs_msg_put(ccs->privfd, CCSPING, -1, -1, 0, NULL) ){
					dbg_printf("connessione non attiva, marca BROKEN\n");
					del_list(TCP_DATA, ccs);
					close(ccs->privfd);
					ccs->privfd = -1;
					if( (gettimeofday(&(ccs->broken_time), NULL)) == -1 ){
						perror("manage_pending_write, gettimeofday");
						exit(-1);
					}
					timerclear(&(ccs->attempt_time));
					if( ptable_isset(ccs, PTCONNECT) ) /* client */
						add_list(PENDING_BROKEN, ccs);
					ccs->priv_status = PRIV_BROKEN;
				} else {
					if( (gettimeofday(&(ccs->read_time), NULL)) == -1 ){
						perror("manage_ccs_ccs_msg, gettimeofday");
						exit(-1);
					}
				}
			}
		}

		ccs = get_next_in_list(MANAGED_CCS, ccs);
	}
}

void manage_pending_close(){
struct ccs_data* ccs;

	ccs = get_list_head(PENDING_CLOSE);
	while ( ccs != NULL ){
		
		del_list(PENDING_CLOSE, ccs);
		del_list(MANAGED_CCS, ccs);

		delete_ccs_data(ccs);

		ccs = get_next_in_list(PENDING_CLOSE, ccs);
	}

	if( to_close && is_list_empty(MANAGED_CCS) ){
		dbg_printf("no more ccs left, exiting\n");
		pthread_exit(NULL);
	}
}


