#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <assert.h>

#include "ccs_mgr_data.h"
#include "ccs_manager.h"

#include "utils.h"
struct ccs_data_list {
	struct ccs_data* ccs;
	struct ccs_data_list* next;
};

static struct ccs_data_list* managed_ccs_list = NULL;
static struct ccs_data_list* pending_accept_list = NULL;
static struct ccs_data_list* pending_connect_list = NULL;
static struct ccs_data_list* pending_write_list = NULL;
static struct ccs_data_list* pending_read_list = NULL;
static struct ccs_data_list* pending_broken_list = NULL; 	
static struct ccs_data_list* pending_close_list = NULL;
static struct ccs_data_list* tcp_listen_list = NULL;
static struct ccs_data_list* tcp_data_list = NULL;

struct ccs_data_list** switch_list(int list_type){
	switch(list_type){
		case MANAGED_CCS:
			return &managed_ccs_list;
		case PENDING_ACCEPT:
			return &pending_accept_list;
		case PENDING_CONNECT:
			return &pending_connect_list;
		case PENDING_WRITE:
			return &pending_write_list;
		case PENDING_READ:
			return &pending_read_list;
		case PENDING_BROKEN:
			return &pending_broken_list;
		case PENDING_CLOSE:
			return &pending_close_list;
		case TCP_LISTEN:
			return &tcp_listen_list;
		case TCP_DATA:
			return &tcp_data_list;
		default:
			dbg_printf("switch_list: wrong list type\n");
			exit(-1);
	}
}

int add_list(int list_type, struct ccs_data* ccs){
struct ccs_data_list** head;
struct ccs_data_list* list;

	if( list_type == PENDING_CLOSE ){
		assert( is_in_list(MANAGED_CCS, ccs) );
		assert( !is_in_list(PENDING_CONNECT, ccs) );
		assert(	!is_in_list(PENDING_WRITE, ccs) );
		assert( !is_in_list(PENDING_READ, ccs) );
		assert( !is_in_list(PENDING_BROKEN, ccs) );
		assert( !is_in_list(TCP_LISTEN, ccs) );
		assert( !is_in_list(TCP_DATA, ccs) );
		dbg_printf("addedd ccs into PENDING_CLOSE\n");
	}

	if( list_type == PENDING_BROKEN ){
		assert( ccs->privfd == -1 );
		dbg_printf("addedd ccs into PENDING_BROKEN\n");
	}
	
	if( list_type == TCP_DATA ){
		assert( ccs->privfd != -1 );
	}

	head = switch_list(list_type);
	list = *head;
	if( *head == NULL ){ /* empty list */
		if( (*head = (struct ccs_data_list*)malloc(sizeof(struct ccs_data_list)))
				== NULL ){
			perror("add_list, malloc");
			exit(-1);
		}
		(*head)->next = NULL;
		(*head)->ccs = ccs;
		ccs->refcount++;
		return 0;
	}
	if (is_in_list(list_type, ccs)) { /* don't duplicate entries */
		dbg_printf("add_list: ccs already in list %d\n", list_type);
		exit(-1);
	}
	while ( list->next != NULL ){ /* go to list end */
		list = list->next;
	}
	if( (list->next = (struct ccs_data_list*)malloc(sizeof(struct ccs_data_list)))
			== NULL ){
		perror("add_list, malloc");
		exit(-1);
	}
	list = list->next;
	list->next = NULL;
	list->ccs = ccs;
	ccs->refcount++;
	return 0;
}

int del_list(int list_type, struct ccs_data* ccs){
struct ccs_data_list** head;
struct ccs_data_list* list;
struct ccs_data_list* previous;

	head = switch_list(list_type);
	list = *head;
	previous = *head;
	while( list != NULL ){
		if( list->ccs == ccs ){
			if( list == *head ){
				*head = list->next;
				free(list);
			} else {
				previous->next = list->next;
				free(list);
			}
			ccs->refcount--;
			return 0;
		}
		previous = list;
		list = list->next;
	}
	dbg_printf("del_list: ccs not in list\n");
	exit(-1);
}

int is_in_list(int list_type, struct ccs_data* ccs){
struct ccs_data_list** head;
struct ccs_data_list* list;

	head = switch_list(list_type);
	list = *head;
	while( list != NULL ){
		if( list->ccs == ccs ){
			return 1;
		}
		list = list->next;
	}
	return 0;
}

int is_list_empty(int list_type){
struct ccs_data_list** head;

	head = switch_list(list_type);

	if( *head == NULL )
		return 1;
	return 0;
}

struct ccs_data* get_list_head(int list_type){
struct ccs_data_list** head;

	head = switch_list(list_type);

	if( *head == NULL ){
		dbg_printf("get_list_head, list is empty\n");
		exit(-1);
	}
	return (*head)->ccs;
}

struct ccs_data* get_next_in_list(int list_type, struct ccs_data* ccs){
struct ccs_data_list** head;
struct ccs_data_list* list;

	head = switch_list(list_type);
	list = *head;
	while( list != NULL ){
		if( list->ccs == ccs ){
            if( list->next == NULL )
                return NULL;

			return list->next->ccs;
		}
		list = list->next;
	}
	return 0;
}

struct ccs_data* get_ccs_from_ccsfd(int ccsfd){
struct ccs_data_list* list;

	list = managed_ccs_list;
	while( list != NULL ){
		if( list->ccs->pubfd == ccsfd ){
			return list->ccs;
		}
		list = list->next;
	}
	return NULL;
}

struct ccs_data* get_ccs_from_mainpubfd(int mainpubfd){
struct ccs_data_list* list;

	list = managed_ccs_list;
	while( list != NULL ){
		if( list->ccs->mainpubfd == mainpubfd ){
			return list->ccs;
		}
		list = list->next;
	}
	return NULL;
}

struct ccs_data* get_ccs_from_privfd(int privfd){
struct ccs_data_list* list;

	list = managed_ccs_list;
	while( list != NULL ){
		if( list->ccs->privfd == privfd ){
			return list->ccs;
		}
		list = list->next;
	}
	return NULL;
}

int empty_list(int list_type){
struct ccs_data_list** head;

	head = switch_list(list_type);
	while( *head != NULL )
		del_list(list_type, (*head)->ccs);
	return 0;
}

int print_list(int list_type){
struct ccs_data_list** head;
struct ccs_data_list* list;

	head = switch_list(list_type);
	list = *head;
	if( list == NULL ){
		printf("list %d is empty\n", list_type);
		return 0;
	}
	printf("ccs_data_list %d: ", list_type);
	do {
		printf("%d ", list->ccs->pubfd);
		list = list->next;
	} while( list != NULL );
	printf("\n");
	return 0;
}

void dump_lists(){
    int i;
    struct ccs_data_list** head;
    struct ccs_data_list* list;
    for(i=0;i<=8;i++){
        head = switch_list(i);
        list = *head;
        printf("%d { ", i);
        
        if( list == NULL ){
            printf("empty } ");
            continue;
        }
       
        do {
            printf("%d ", list->ccs->pubfd);
            list = list->next;
        } while( list != NULL );
        
        printf("} ");
    }
    printf("\n");
    fflush(NULL);
}


struct ccs_data* new_ccs_data(void){
struct ccs_data* ccs = NULL;
int* sp = NULL;
struct linger linopt;
socklen_t optlen = sizeof(struct linger);


	if( !(ccs = (struct ccs_data*)malloc(sizeof(struct ccs_data))) ){
		perror("malloc for ccs_data");
		exit(-1);
	}
	if( !(sp=malloc(2*sizeof(int))) ){
		perror("alloc socketpair");
		exit(-1);
	}
	if((socketpair(AF_UNIX,SOCK_STREAM,0, sp)) != 0){
		perror("socketpair");
		exit(-1);
	}
	linopt.l_onoff = 1;
	linopt.l_linger = LINGER_TIMEOUT;
	if( setsockopt(sp[0], SOL_SOCKET, SO_LINGER, &linopt, optlen) ){
		perror("setting linger on socketpair\n");
		exit(-1);
	}
	if( setsockopt(sp[1], SOL_SOCKET, SO_LINGER, &linopt, optlen) ){
		perror("setting linger on socketpair\n");
		exit(-1);
	}

	ccs->pubfd = sp[1];
	ccs->mainpubfd = sp[0];
	free(sp);
	ccs->remfd = -1;
	ccs->privfd = -1;
	ccs->father = NULL;
	
	ccs->wbuf = init_buffer();
	ccs->rbuf = init_buffer();
	
	ccs->main_rbuf = NULL;
	ccs->main_rbufsize = 0;
	ccs->main_acceptaddr = NULL;
	
	ccs->ptable[PTRADDR] = NULL;
	ccs->ptable[PTBIND] = NULL;
	ccs->ptable[PTLISTEN] = NULL;
	ccs->ptable[PTACCEPT] = NULL;
	ccs->ptable[PTCONNECT] = NULL;

	ccs->pub_status = PUB_UNUSED;
	ccs->priv_status = PRIV_UNUSED;

	ccs->refcount = 0;
	ccs->conn_cache = 0;

	timerclear(&(ccs->broken_time));
	timerclear(&(ccs->attempt_time));
	timerclear(&(ccs->read_time));
	
	return ccs;
}

int delete_ccs_data(struct ccs_data* ccs){
int tmpbuf;
size_t bufsz = sizeof(int);

	if(ccs->refcount){
		dbg_printf("ERROR: trying to delete ccs still in use\n");
		exit(-1);
	}
	close(ccs->pubfd);
	close(ccs->mainpubfd);
	if(ccs->privfd != -1){
		shutdown(ccs->privfd,SHUT_WR);
		read(ccs->privfd, &tmpbuf, bufsz);
	}
	destroy_buffer(ccs->wbuf);
	destroy_buffer(ccs->rbuf);

	if(ptable_isset(ccs, PTRADDR))
		ptable_unset(ccs, PTRADDR);
	if(ptable_isset(ccs, PTBIND))
		ptable_unset(ccs, PTBIND);
	if(ptable_isset(ccs, PTLISTEN))
		ptable_unset(ccs, PTLISTEN);
	if(ptable_isset(ccs, PTACCEPT))
		ptable_unset(ccs, PTACCEPT);
	if(ptable_isset(ccs, PTCONNECT))
		ptable_unset(ccs, PTCONNECT);
	free(ccs);
	return 1;
}


int ptable_isset(struct ccs_data* ccs, int pttype){

	switch(pttype){
		case PTRADDR:
		case PTBIND:
		case PTLISTEN:
		case PTACCEPT:
		case PTCONNECT:
			if ( ccs->ptable[pttype] == NULL ){
				return 0;
			} else {
				return 1;
			}
		default:
			dbg_printf("ptable_isset, wrong table entry\n");
			exit(-1);
	}
}

void ptable_unset(struct ccs_data* ccs, int pttype){

	switch(pttype){
		case PTRADDR:
		case PTBIND:
		case PTLISTEN:
		case PTACCEPT:
		case PTCONNECT:
			free(ccs->ptable[pttype]);
			ccs->ptable[pttype] = NULL;
			break;
		default:
			dbg_printf("ptable_unset, wrong table entry\n");
			exit(-1);
	}
	
}

void ptable_get_int(struct ccs_data* ccs, int pttype, int* dst){

	if( !ptable_isset(ccs, pttype) ){
		dbg_printf("ptable_get_int, type not set\n");
		exit(-1);
	}
	if( memcpy(dst, ccs->ptable[pttype], sizeof(int)) != dst ){
		perror("ptable_get_int, memcpy");
		exit(-1);
	}
}

void ptable_set_int(struct ccs_data* ccs, int pttype, int* src){
	
	if( ptable_isset(ccs, pttype) ){
		dbg_printf("ptable_set_int, type already set\n");
		exit(-1);
	}
	if( (ccs->ptable[pttype] = malloc(sizeof(int))) == NULL ){
		perror("ptable_set_int, malloc");
		exit(-1);
	}
	if( memcpy(ccs->ptable[pttype], src, sizeof(int)) != ccs->ptable[pttype] ){
		perror("ptable_set_int, memcpy");
		exit(-1);
	}
}

void ptable_get_addr(struct ccs_data* ccs, int pttype, struct sockaddr* dst){
	
	if( !ptable_isset(ccs, pttype) ){
		dbg_printf("ptable_get_addr, type not set\n");
		exit(-1);
	}
	if( memcpy(dst, ccs->ptable[pttype], sizeof(struct sockaddr)) != dst ){
		perror("ptable_get_addr, memcpy");
		exit(-1);
	}
}

void ptable_set_addr(struct ccs_data* ccs, int pttype, struct sockaddr* src){
	
	if( ptable_isset(ccs, pttype) ){
		dbg_printf("ptable_set_addr, type already set\n");
		exit(-1);
	}
	if( (ccs->ptable[pttype] = malloc(sizeof(struct sockaddr))) == NULL ){
		perror("ptable_set_addr, malloc");
		exit(-1);
	}
	if( memcpy(ccs->ptable[pttype], src, sizeof(struct sockaddr))
			!= ccs->ptable[pttype] ){
		perror("ptable_set_addr, memcpy");
		exit(-1);
	}
}


/* TODO: modificare le funzioni per uscire con errore se il buffer
 * e' pieno
 */
struct ccs_buf* init_buffer(){
    
    struct ccs_buf *buf;

	if( !(buf =(struct ccs_buf*)malloc(sizeof(struct ccs_buf))) ){
		perror("init_buffer, malloc struct");
		exit(-1);
	}
	buf->avail = 0;
	if( !(buf->head =(char *)malloc(sizeof(char)*BUF_MAXLEN)) ){
		perror("init_buffer, malloc buffer");
		exit(-1);
	}
	if( memset(buf->head, 0, sizeof(char)*BUF_MAXLEN) != buf->head ){
		perror("init_buffer, memset");
		exit(-1);
	}
	
    return buf;
}

void destroy_buffer(struct ccs_buf* buf){

	if(buf == NULL){
		dbg_printf("destroy_buffer, buf no good\n");
		exit(-1);
	}
	free(buf->head);
	free(buf);
}

int copy_to_buffer(struct ccs_buf* buf, void* data, size_t len){
char* bufptr = buf->head;

    assert(len > 0);
    assert(data > 0);

	if( len + buf->avail > BUF_MAXLEN ){
		dbg_printf("copy_to_buffer, size exceeded\n");
		return 0;
	}
	bufptr = buf->head + buf->avail;
	if( memcpy(bufptr, data, len) != bufptr ){
		perror("copy_to_buffer, memcpy");
		exit(-1);
	}
	buf->avail = buf->avail + len;

	return 1;
}

int copy_from_buffer(struct ccs_buf* buf, void* data, size_t len){
char* bufptr = buf->head;
    
    assert(len > 0);
    assert(data > 0);

	if( len > buf->avail ){
		dbg_printf("copy_from_buffer, not enought data\n");
		return 0;
	}
	if( memcpy(data, bufptr, len) != data ){
		perror("copy_from_buffer, memcpy");
		exit(-1);
	}
	buf->avail = buf->avail - len;
	bufptr = buf->head + len;
	if( memmove(buf->head, bufptr, buf->avail) != buf->head ){
		perror("copy_from_buffer, memmove");
		exit(-1);
	}
	bufptr = buf->head + buf->avail;
	if( memset(bufptr, 0, BUF_MAXLEN - buf->avail) != bufptr ){
		perror("copy_from_buffer, memset");
		exit(-1);
	}

	return 1;
}

int del_from_buffer(struct ccs_buf* buf, size_t len){
char* bufptr = buf->head;

    assert(len > 0);

	if( len > buf->avail ){
		dbg_printf("del_from_buffer, not enought data\n");
		return 0;
	}
	buf->avail = buf->avail - len;
	bufptr = buf->head + len;
	if( memmove(buf->head, bufptr, buf->avail) != buf->head ){
		perror("del_from_buffer, memmove");
		exit(-1);
	}
	bufptr = buf->head + buf->avail;
	if( memset(bufptr, 0, BUF_MAXLEN - buf->avail) != bufptr ){
		perror("del_from_buffer, memset");
		exit(-1);
	}

	return 1;
}


