
#ifndef __CCS_MGR_DATA_H__
#define __CCS_MGR_DATA_H__

#include <unistd.h>
#include <sys/time.h>

/* History table entries */
#define PTRADDR	0
#define PTBIND 1
#define PTLISTEN 2
#define PTACCEPT 3
#define PTCONNECT 4

/* Public status */
#define PUB_UNUSED -1
#define PUB_SOCKET 0
#define PUB_BIND 1
#define PUB_LISTEN 2
#define PUB_ACTIVE 3
#define PUB_CLOSE 4

/* Private status */
#define PRIV_UNUSED -1
#define PRIV_SOCKET 0
#define PRIV_BIND 1
#define PRIV_LISTEN 2
#define PRIV_ACTIVE 3
#define PRIV_BROKEN 4

/* Data used by every ccs */

#define BUF_MAXLEN (3 * 1024 * 1024)
#define MAXDATALEN 4096

struct ccs_buf {
	size_t avail; /* amount of data available */
	char* head; /* head of buffer */
};

struct ccs_data {
	int pubfd; /* public ccsfd */
	int mainpubfd; /* public fd, main side */
	int remfd; /* public ccsfd of remote part */
	int privfd; /* private fd */
	
	struct ccs_data* father; /* if created from an accept this is the listening ccs */
	
	struct ccs_buf* wbuf; /* write buffer */
	struct ccs_buf* rbuf; /* read buffer */
	
	void* main_rbuf; /* pointer to main read buffer */
	size_t main_rbufsize; /* size of main read buffer */
	struct sockaddr* main_acceptaddr; /* pointer to main accept sockaddr */
	
	void* ptable[5]; /* history table for parameters set onto ccs:
			  reuseaddr, bind, listen, accept, connect */
	int pub_status; /* status for main */
	int priv_status; /* internal status */
	
	int refcount; /* reference counter, used to know if the ccs appears
			 into a pending_* list */
	
	int conn_cache; /* number of pending_connect not accepted:
			   must be <= ptable[PTLISTEN]->backlog */
	
	struct timeval broken_time; /* "when" connection is broken */
	struct timeval attempt_time; /* last attempt for connection resume */
	struct timeval read_time; /* last data received from a ccs in pending_read */
};

/* ccs lists type */
#define MANAGED_CCS 0 		/* every existing ccs should stay here */
#define PENDING_ACCEPT 1	/* ccs that have performed an accept */
#define PENDING_CONNECT 2	/* ccs awaiting for an accept */
#define PENDING_WRITE 3		/* ccs that have to send data */
#define PENDING_READ 4		/* ccs that have performed a read */
#define PENDING_BROKEN 5	/* ccs that have to be repristinated (only client) */
#define PENDING_CLOSE 6		/* ccs that we must close */
#define TCP_LISTEN 7		/* listening ccs */
#define TCP_DATA 8		/* ccs that can read/write data */

/* ccs list managing functions */
int add_list(int list_type, struct ccs_data* ccs);
int del_list(int list_type, struct ccs_data* ccs);
int is_in_list(int list_type, struct ccs_data* ccs);
int is_list_empty(int list_type);
struct ccs_data* get_list_head(int list_type);
struct ccs_data* get_next_in_list(int list_type, struct ccs_data* ccs);
struct ccs_data* get_ccs_from_ccsfd(int ccsfd);
struct ccs_data* get_ccs_from_mainpubfd(int mainpubfd);
struct ccs_data* get_ccs_from_privfd(int privfd);
int empty_list(int list_type);
int print_list(int list_type);

/* ccs_data utilities */
struct ccs_data* new_ccs_data();
int delete_ccs_data(struct ccs_data* ccs); 

int  ptable_isset(struct ccs_data* ccs, int pttype);
void ptable_unset(struct ccs_data* ccs, int pttype);
void ptable_get_int(struct ccs_data* ccs, int pttype, int* dst);
void ptable_set_int(struct ccs_data* ccs, int pttype, int* src);
void ptable_get_addr(struct ccs_data* ccs, int pttype, struct sockaddr* dst);
void ptable_set_addr(struct ccs_data* ccs, int pttype, struct sockaddr* src);

struct ccs_buf* init_buffer();
void destroy_buffer(struct ccs_buf* buf);
int copy_to_buffer(struct ccs_buf* buf, void* data, size_t len);
int copy_from_buffer(struct ccs_buf* buf, void* data, size_t len);
int del_from_buffer(struct ccs_buf* buf, size_t len);

#endif /* ccs_mgr_data.h */
