#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <arpa/inet.h>
#include <assert.h>

#include "interproto.h"

#include "utils.h"

#define PROTOACK 42


int send_field(int fd, void* buf, size_t size){
fd_set set;
struct timeval timeout;
int tmpmsg;

	FD_ZERO(&set);
	FD_SET(fd, &set);
	timeout.tv_sec = INACTIVITY_TIMEOUT_SECONDS;
	timeout.tv_usec = 0;
	
	if( !select(fd+1, NULL, &set, NULL, &timeout) ){
		return 0;
	}
	if( write(fd, buf, size) != size ){
		return 0;
	}

	/* get protocol ack */
	FD_ZERO(&set);
	FD_SET(fd, &set);
	timeout.tv_sec = INACTIVITY_TIMEOUT_SECONDS;
	timeout.tv_usec = 0;
	
	if( !select(fd+1, &set, NULL, NULL, &timeout) ){
		return 0;
	}
	if( read(fd, &tmpmsg, sizeof(int)) != sizeof(int) ){
		return 0;
	}

	return 1;
}

int recv_field(int fd, void* buf, size_t size){
fd_set set;
struct timeval timeout;
int tmpmsg;

	FD_ZERO(&set);
	FD_SET(fd, &set);
	timeout.tv_sec = INACTIVITY_TIMEOUT_SECONDS;
	timeout.tv_usec = 0;
	
	if( !select(fd+1, &set, NULL, NULL, &timeout) ){
		return 0;
	}
	if( read(fd, buf, size) != size ){
		return 0;
	}

	/* put protocol ack */
	FD_ZERO(&set);
	FD_SET(fd, &set);
	timeout.tv_sec = INACTIVITY_TIMEOUT_SECONDS;
	timeout.tv_usec = 0;
	tmpmsg = PROTOACK;

	if( !select(fd+1, NULL, &set, NULL, &timeout) ){
		return 0;
	}
	if( write(fd, &tmpmsg, sizeof(int)) != sizeof(int) ){
		return 0;
	}


	return 1;

}

struct ccs_msg* ccs_msg_get(int fd){
struct ccs_msg* msg;

	entra();

	if( (msg=(struct ccs_msg*)malloc(sizeof(struct ccs_msg))) == NULL ){
		perror("ccs_msg_get, malloc");
		exit(-1);
	}

	/* get msg->type */
	if( !recv_field(fd, &(msg->type), sizeof(int)) ){
		dbg_printf("connection timeout\n");
		free(msg);
		return NULL;
	}
	msg->type = ntohl(msg->type);

	/* get msg->pubfd */
	if( !recv_field(fd, &(msg->pubfd), sizeof(int)) ){
		dbg_printf("connection timeout\n");
		free(msg);
		return NULL;
	}
	msg->pubfd = ntohl(msg->pubfd);
	
	/* get msg->rem_pubfd */
	if( !recv_field(fd, &(msg->rem_pubfd), sizeof(int)) ){
		dbg_printf("connection timeout\n");
		free(msg);
		return NULL;
	}
	msg->rem_pubfd = ntohl(msg->rem_pubfd);
	
	/* if data we have to get payload */
	if( msg->type == CCSDATA ){

		/* get msg->bodysize */
		if( !recv_field(fd, &(msg->bodysize), sizeof(int)) ){
			dbg_printf("connection timeout\n");
			free(msg);
			return NULL;
		}
		msg->bodysize = ntohl(msg->bodysize);
		dbg_printf("da leggere: %d\n", msg->bodysize);

		if( (msg->body=malloc(sizeof(char)*msg->bodysize)) == NULL ){
			perror("ccs_msg_get, malloc body");
			free(msg);
			exit(-1);
		}

		/* get msg->body */
		if( !recv_field(fd, msg->body, msg->bodysize) ){
			dbg_printf("connection timeout\n");
			free(msg->body);
			free(msg);
			return NULL;
		}
		
	} else {
		msg->bodysize = 0;
		msg->body = NULL;
	}

	esci();
	return msg;
}

int ccs_msg_put(int fd, int type, int pubfd, int rem_pubfd, size_t size, void* buf){

	entra();
	assert( type >= -1 && type <= 6 );
    
	if( type == CCSDATA )
		assert( size > 0 && buf > 0 && pubfd >= 0);
    
	/* put msg->type */
	type = htonl(type);
	if( !send_field(fd, &type, sizeof(int)) ){
		dbg_printf("connection timeout\n");
		return 0;
	}
	type = ntohl(type);

	/* put msg->pubfd */
	/* we exchange pubfd and rem_pubfd order for peer */
	rem_pubfd = htonl(rem_pubfd);
	if( !send_field(fd, &rem_pubfd, sizeof(int)) ){
		dbg_printf("connection timeout\n");
		return 0;
	}

	/* put msg->rem_pubfd */
	pubfd = htonl(pubfd);
	if( !send_field(fd, &pubfd, sizeof(int)) ){
		dbg_printf("connection timeout\n");
		return 0;
	}

	/* if data we have to send payload */
	if( type == CCSDATA ){

		/* put msg->bodysize */
		size = htonl(size);
		if( !send_field(fd, &size, sizeof(int)) ){
			dbg_printf("connection timeout\n");
			return 0;
		}
		size = ntohl(size);

		/* put msg->body */
		if( !send_field(fd, buf, size) ){
			dbg_printf("connection timeout\n");
			return 0;
		}

	}
	
	esci();
	return 1;

}

void ccs_msg_del(struct ccs_msg *message){

	if( message->type == CCSDATA )
		free(message->body);
	free(message);

}
