#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <pthread.h>

#include "main_data.h"

#include "utils.h"

struct ccsfd{
	struct ccsfd*	next;
	pthread_mutex_t fdmux;
	int		fd;
};

static struct ccsfd* head = NULL;
static pthread_mutex_t ccsfdmux;

int initccslist(){
	
	if(pthread_mutex_init(&ccsfdmux, NULL)){
		perror("initccslist");
		exit(-1);
	}

	return 0;
}

int destroyccslist(){

	while( head != NULL )
		delccslist(head->fd);

	if( pthread_mutex_unlock(&ccsfdmux) ){
		perror("destroyccslist, unlock mutex");
		exit(-1);
	}
	if( pthread_mutex_destroy(&ccsfdmux) ){
		perror("destroyccslist, destroy mutex");
	}
	return 0;
}

int acquireccslist(){

	if(pthread_mutex_lock(&ccsfdmux)){
		perror("acquireccslist");
		exit(-1);
	}
	return 0;
}

int releaseccslist(){
	
	if( pthread_mutex_unlock(&ccsfdmux) ){
		perror("releaseccslist");
		exit(-1);
	}
	return 0;
}

pthread_mutex_t* isinccslist(int fd){
struct ccsfd* list = head;

	while( list != NULL ){
		if( list->fd == fd ){
			return &(list->fdmux);
		}
		list = list->next;
	}
	return NULL;
}

int addccslist(int fd){
struct ccsfd* list = head;

	if( head == NULL ){ /* empty list */
		if( (head = (struct ccsfd*)malloc(sizeof(struct ccsfd)))
				== NULL ){
			perror("addccslist, malloc");
			exit(-1);
		}
		head->next = NULL;
		pthread_mutex_init(&(head->fdmux), NULL);
		head->fd = fd;
		return 0;
	}
	
	if ( isinccslist(fd) != NULL ) { /* don't duplicate entries */
		dbg_printf("addccslist: fd already in list\n");
		exit(-1);
	}

	while ( list->next != NULL ){ /* go to list end */
		list = list->next;
	}

	if( (list->next = (struct ccsfd*)malloc(sizeof(struct ccsfd))) == NULL ){
		perror("addccslist, malloc");
		exit(-1);
	}
	list = list->next;
	list->next = NULL;
	pthread_mutex_init(&(head->fdmux), NULL);
	list->fd = fd;
	return 0;
}

int delccslist(int fd){
struct ccsfd* list = head;
struct ccsfd* previous = head;

	while( list != NULL ){
		if( list->fd == fd ){
			pthread_mutex_destroy(&(list->fdmux));
			if( list == head ){
				head = list->next;
				free(list);
			} else {
				previous->next = list->next;
				free(list);
			}
			return 0;
		}
		previous = list;
		list = list->next;
	}
	dbg_printf("delccslist: fd not in list\n");
	exit(-1);
}
	
int printccslist(){
struct ccsfd* list = head;

	if( list == NULL ){
		dbg_printf("list is empty\n");
		return 0;
	}
	dbg_printf("ccslist: ");
	do {
		dbg_printf("%d ", list->fd);
		list = list->next;
	} while( list != NULL );
	dbg_printf("\n");
	return 0;
}


